unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, XPMan, StdCtrls, JPEG, Spin, Wait, RXSpin;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    btnLoadImg: TButton;
    XPManifest1: TXPManifest;
    btnApplyNoiseReductionCA: TButton;
    btnApplyEdgeDetectionCA: TButton;
    Panel2: TPanel;
    Panel3: TPanel;
    Splitter1: TSplitter;
    ScrollBox1: TScrollBox;
    OriginalImg: TImage;
    ScrollBox2: TScrollBox;
    ProcessedImg: TImage;
    btnSaveProcessedImg: TButton;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    Panel4: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    NoiseSteps: TSpinEdit;
    chkShowProgress: TCheckBox;
    chkShowEdge: TCheckBox;
    Label4: TLabel;
    Epsilon: TRxSpinEdit;
    Label5: TLabel;
    EdgeSteps: TSpinEdit;
    procedure btnLoadImgClick(Sender: TObject);
    procedure btnSaveProcessedImgClick(Sender: TObject);
    procedure btnApplyNoiseReductionCAClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnApplyEdgeDetectionCAClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure NoiseReductionCA(Step: Integer; Bitmap: TBitmap; Output: TBitmap);
    procedure EdgeDetectionCA(Step: Integer; Bitmap: TBitmap; Output: TBitmap);
  end;

var
  MainForm: TMainForm;
  Wait: TWaitForm;
  Iterations: Longint;

implementation

{$R *.dfm}

procedure TMainForm.btnLoadImgClick(Sender: TObject);
var JPG: TJPEGImage;
begin
  if OpenDialog.Execute then
  begin
    if LowerCase(ExtractFileExt(OpenDialog.FileName))='.bmp' then
     OriginalImg.Picture.LoadFromFile(OpenDialog.FileName)
    else
    begin
      // load a JPEG file from disk
      JPG := TJPEGImage.Create;
      JPG.LoadFromFile(OpenDialog.FileName);
      OriginalImg.Picture.Bitmap.Assign(JPG);
      JPG.Free;
    end;
    if (OriginalImg.Picture.Bitmap.PixelFormat <> pf1bit) and
       (OriginalImg.Picture.Bitmap.PixelFormat <> pf4bit) and
       (OriginalImg.Picture.Bitmap.PixelFormat <> pf8bit) then
    begin
      OriginalImg.Picture.Bitmap.PixelFormat := pf8Bit;
      ShowMessage('Your image''s quality has been automatically reduced to 8 bits per pixels (256 colors).');
    end;
    OriginalImg.Width := OriginalImg.Picture.Bitmap.Width;
    OriginalImg.Height := OriginalImg.Picture.Bitmap.Height;
  end;
end;

procedure TMainForm.btnSaveProcessedImgClick(Sender: TObject);
var JPG: TJPEGImage;
begin
  if ProcessedImg.Picture = nil then
  begin
    ShowMessage('Your processed image is empty!');
    Exit;
  end;

  if SaveDialog.Execute then
  begin
    if SaveDialog.FilterIndex = 1 then // BMP
     ProcessedImg.Picture.Bitmap.SaveToFile(SaveDialog.FileName)
    else
    begin
      // save JPEG to disk
      JPG := TJPEGImage.Create;
      JPG.Assign(ProcessedImg.Picture.Bitmap);
      JPG.SaveToFile(SaveDialog.FileName);
      JPG.Free;
    end;
  end;
end;

procedure TMainForm.btnApplyNoiseReductionCAClick(Sender: TObject);
var Bitmap, Outbitmap: TBitmap;
    I: Integer;
begin
  ProcessedImg.Width := OriginalImg.Picture.Bitmap.Width;
  ProcessedImg.Height := OriginalImg.Picture.Bitmap.Height;
  Enabled := False;
  Wait.PBar.Position := 0;
  Wait.Show;
  Iterations := 0;
  if NoiseSteps.Value = 1 then
  begin
    if chkShowProgress.Checked then
     NoiseReductionCA(1, OriginalImg.Picture.Bitmap, ProcessedImg.Picture.Bitmap)
    else
    begin
      Outbitmap := TBitmap.Create;
      Outbitmap.Assign(OriginalImg.Picture.Bitmap);
      Wait.lblStatus.Caption := 'Please wait, processing image...'; 
      NoiseReductionCA(1, OriginalImg.Picture.Bitmap, OutBitmap);
      ProcessedImg.Picture.Bitmap.Assign(OutBitmap);
      OutBitmap.Free;
    end;
  end
  else
  begin
    Bitmap := TBitmap.Create;
    Bitmap.Assign(OriginalImg.Picture.Bitmap);
    if not chkShowProgress.Checked then
    begin
      Outbitmap := TBitmap.Create;
      Outbitmap.Assign(Bitmap);
    end;
    for I:=1 to NoiseSteps.Value do
    begin
      Wait.lblStatus.Caption := 'Performing step '+IntToStr(I)+' of '+IntToStr(NoiseSteps.Value)+'...'; 
      Iterations := 0;
      if chkShowProgress.Checked then
      begin
        NoiseReductionCA(I, Bitmap, ProcessedImg.Picture.Bitmap);
        Bitmap.Assign(ProcessedImg.Picture.Bitmap);
      end
      else
      begin
        NoiseReductionCA(I, Bitmap, OutBitmap);
        Bitmap.Assign(OutBitmap);
      end;
    end;
    Bitmap.Free;
    OutBitmap.Free;
  end;
  Wait.Hide;
  Enabled := True;
end;

procedure TMainForm.NoiseReductionCA(Step: Integer; Bitmap: TBitmap; Output: TBitmap);

function GetMax(List: TStringList): TColor;
var I, J, Max, MaxPos: Integer;
    Color: TColor;
    MaxList: array of Integer;
begin
  // make a vector for each element with the number of appearances
  SetLength(MaxList, List.Count);
  for I:=0 to List.Count-1 do
  begin
    Max := 0;
    for J:=0 to List.Count-1 do
     if List[I] = List[J] then
      Inc(Max);
    MaxList[I] := Max;
  end;

  // get the maximum of the vector MaxList
  Max := MaxList[0];
  MaxPos := 0;
  for I:=1 to Length(MaxList)-1 do
   if MaxList[I] > Max then
   begin
     Max := MaxList[I];
     MaxPos := I;
   end;

  // return the color here
  Result := StrToInt(List[MaxPos]);
  SetLength(MaxList, 0);
end;

var I, J: Integer;
    List: TStringList;
begin
  List := TStringList.Create;
  ProcessedImg.Picture.Bitmap.Assign(Bitmap);
  for i := 0 to Bitmap.Width-1 do
   for j := 0 to Bitmap.Height-1 do
   begin
     List.Clear;

     if I > 0 then
      List.Add(IntToStr(Bitmap.Canvas.Pixels[I-1, J]));
     if J > 0 then
      List.Add(IntToStr(Bitmap.Canvas.Pixels[I, J-1]));
     if I < Bitmap.Width-1 then
      List.Add(IntToStr(Bitmap.Canvas.Pixels[I+1, J]));
     if J < Bitmap.Height-1 then
      List.Add(IntToStr(Bitmap.Canvas.Pixels[I, J+1]));
     List.Add(IntToStr(Bitmap.Canvas.Pixels[I, J]));

     Output.Canvas.Pixels[I, J] := GetMax(List);

     // update GUI here
     Application.ProcessMessages;
     Inc(Iterations);
     Wait.PBar.Position := (100 * Iterations) div (Bitmap.Width * Bitmap.Height);
   end;
  List.Free;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Wait := TWaitForm.Create(Self);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Wait.Free;
end;

procedure TMainForm.btnApplyEdgeDetectionCAClick(Sender: TObject);
var Bitmap, Outbitmap: TBitmap;
    I: Integer;
begin
  ProcessedImg.Width := OriginalImg.Picture.Bitmap.Width;
  ProcessedImg.Height := OriginalImg.Picture.Bitmap.Height;
  Enabled := False;
  Wait.PBar.Position := 0;
  Wait.Show;
  Iterations := 0;
  if EdgeSteps.Value = 1 then
  begin
    if chkShowEdge.Checked then
     EdgeDetectionCA(1, OriginalImg.Picture.Bitmap, ProcessedImg.Picture.Bitmap)
    else
    begin
      Outbitmap := TBitmap.Create;
      Outbitmap.Assign(OriginalImg.Picture.Bitmap);
      Wait.lblStatus.Caption := 'Please wait, processing image...';
      EdgeDetectionCA(1, OriginalImg.Picture.Bitmap, OutBitmap);
      ProcessedImg.Picture.Bitmap.Assign(OutBitmap);
      OutBitmap.Free;
    end;
  end
  else
  begin
    Bitmap := TBitmap.Create;
    Bitmap.Assign(OriginalImg.Picture.Bitmap);
    if not chkShowEdge.Checked then
    begin
      Outbitmap := TBitmap.Create;
      Outbitmap.Assign(Bitmap);
    end;
    for I:=1 to EdgeSteps.Value do
    begin
      Wait.lblStatus.Caption := 'Performing step '+IntToStr(I)+' of '+IntToStr(EdgeSteps.Value)+'...';
      Iterations := 0;
      if chkShowEdge.Checked then
      begin
        EdgeDetectionCA(I, Bitmap, ProcessedImg.Picture.Bitmap);
        Bitmap.Assign(ProcessedImg.Picture.Bitmap);
      end
      else
      begin
        EdgeDetectionCA(I, Bitmap, OutBitmap);
        Bitmap.Assign(OutBitmap);
      end;
    end;
    Bitmap.Free;
    OutBitmap.Free;
  end;
  Wait.Hide;
  Enabled := True;
end;

procedure TMainForm.EdgeDetectionCA(Step: Integer; Bitmap: TBitmap; Output: TBitmap);

function GetNrOfColors(PixelFormat: TPixelFormat): Integer;
begin
  if PixelFormat = pf1Bit then
   Result := 2
  else
  if PixelFormat = pf4Bit then
   Result := 16
  else
   Result := 256;
end;

var I, J: Integer;
    IsNull: Boolean;
    Pal : array[0..255] of PaletteEntry; // we'll be needing the palette for this
begin
  ProcessedImg.Picture.Bitmap.Assign(Bitmap);
  GetPaletteEntries(Bitmap.palette, 0, GetNrOfColors(Bitmap.PixelFormat), Pal); {save palette into pal}
  for i := 0 to Bitmap.Width-1 do
   for j := 0 to Bitmap.Height-1 do
   begin

     IsNull := True;

     // borders?
     if (I = 0) or (I = Bitmap.Width - 1) or (J = 0) or (J = Bitmap.Height - 1) then
      if Abs(Bitmap.Canvas.Pixels[I, J] - RGB(Pal[0].peRed, Pal[0].peGreen, Pal[0].peBlue)) >= Epsilon.Value then
       IsNull := False;

     // not borders, need to test
     if I > 0 then
      if Abs(Bitmap.Canvas.Pixels[I-1, J]-Bitmap.Canvas.Pixels[I, J]) >= Epsilon.Value then
       IsNull := False;

     if I < Bitmap.Width-1 then
      if Abs(Bitmap.Canvas.Pixels[I+1, J]-Bitmap.Canvas.Pixels[I, J]) >= Epsilon.Value then
       IsNull := False;

     if J > 0 then
      if Abs(Bitmap.Canvas.Pixels[I, J-1]-Bitmap.Canvas.Pixels[I, J]) >= Epsilon.Value then
       IsNull := False;

     if J < Bitmap.Height-1 then
      if Abs(Bitmap.Canvas.Pixels[I, J+1]-Bitmap.Canvas.Pixels[I, J]) >= Epsilon.Value then
       IsNull := False;

     if IsNull then
      Output.Canvas.Pixels[I, J] := RGB(Pal[0].peRed, Pal[0].peGreen, Pal[0].peBlue); // first color in the palette

     // update GUI here
     Application.ProcessMessages;
     Inc(Iterations);
     Wait.PBar.Position := (100 * Iterations) div (Bitmap.Width * Bitmap.Height);
   end;
end;

end.
