unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Spin, XPMan, UnitBitmapResize;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    ScrollBox: TScrollBox;
    Label1: TLabel;
    XPManifest1: TXPManifest;
    ImgWidth: TSpinEdit;
    Label2: TLabel;
    ImgHeight: TSpinEdit;
    btnBringToLife: TButton;
    Image: TImage;
    Label3: TLabel;
    DigitalZoom: TSpinEdit;
    Label4: TLabel;
    PauseEdit: TSpinEdit;
    Label5: TLabel;
    Dispersion: TSpinEdit;
    Bevel1: TBevel;
    Bevel2: TBevel;
    btnClose: TButton;
    procedure btnBringToLifeClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
  public
    { Public declarations }
    StopAnimation: Boolean;
    MustClose: Boolean;
    MyVector, Tmp: array[0..1001, 0..1001] of TColor;
  end;

var
  MainForm: TMainForm;
  Bitmap: TBitmap;
  WHITE, BLACK: TColor;

implementation

{$R *.dfm}

procedure TMainForm.btnBringToLifeClick(Sender: TObject);

function GetRandom(N: Integer): TColor;
begin
  if N = 0 then
   Result := WHITE
  else
   Result := BLACK;
end;

function CountAlive(I, J: Integer; Original: TColor): TColor;
var N: Integer;
begin
  N := 0;
  if MyVector[I-1, J-1] = BLACK then
   Inc(N);
  if MyVector[I, J-1] = BLACK then
   Inc(N);
  if MyVector[I, J+1] = BLACK then
   Inc(N);
  if MyVector[I-1, J] = BLACK then
   Inc(N);
  if MyVector[I+1, J] = BLACK then
   Inc(N);
  if MyVector[I+1, J-1] = BLACK then
   Inc(N);
  if MyVector[I+1, J] = BLACK then
   Inc(N);
  if MyVector[I+1, J+1] = BLACK then
   Inc(N);

  Result := WHITE; // assume dead cell

  if Original = WHITE then // dead cell with exactly 3 live neighbours
   if (N = 2) or (N = 3) then
    Result := BLACK; // comes alive...

  if Original = BLACK then // a live cell with 2 or 3 live neighbours?
   if (N < 2) or (N > 3) then
    Result := BLACK; // comes alive...
end;

var I, J, K, L: Integer;
begin
  if btnBringToLife.Caption = 'Stop Animation' then
  begin
    btnBringToLife.Enabled := False;
    StopAnimation := True;
    Exit;
  end;

  StopAnimation := False;
  Bitmap.Width := ImgWidth.Value;
  Bitmap.Height := ImgWidth.Value;
  Bitmap.PixelFormat := pf1Bit;
  { Initialize image with random data }
  for I:=0 to 1001 do
   for J:=0 to 1001 do
    MyVector[I, J] := WHITE;
  Randomize;
  L := 0;
  for I:=0 to Bitmap.Width-1 do
   for J:=0 to Bitmap.Height-1 do
   begin
     Inc(L);
     if L mod (100 div Dispersion.Value) = 0 then
     begin
       K := Random(2);
       Bitmap.Canvas.Pixels[I, J] := GetRandom(K);
       MyVector[I+1, J+1] := GetRandom(K);
     end
     else
     begin
       Bitmap.Canvas.Pixels[I, J] := WHITE;
       MyVector[I+1, J+1] := WHITE;
     end;
   end;

  BitmapResize(Bitmap, Image.Picture.Bitmap, DigitalZoom.Value);
  Image.Width := Image.Picture.Bitmap.Width;
  Image.Height := Image.Picture.Bitmap.Height;

  btnBringToLife.Caption := 'Stop Animation';

  repeat
    Application.ProcessMessages;

    { Simulate game of life }
    Tmp := MyVector;
    for I:=1 to Bitmap.Width do
     for J:=1 to Bitmap.Height do
     begin
       Tmp[I, J] := CountAlive(I, J, MyVector[I, J]);
       Bitmap.Canvas.Pixels[I-1, J-1] := Tmp[I, J];
     end;
    MyVector := Tmp;

    { Redraw animation }
    BitmapResize(Bitmap, Image.Picture.Bitmap, DigitalZoom.Value);

    { Pause }
    Sleep(PauseEdit.Value);
  until StopAnimation;
  btnBringToLife.Caption := 'Bring To Life!';
  btnBringToLife.Enabled := True;
  if MustClose then
   Close;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  StopAnimation := False;
  Bitmap := TBitmap.Create;
  WHITE := RGB(255, 255, 255);
  BLACK := RGB(0, 0, 0);
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Bitmap.Free;
end;

procedure TMainForm.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if btnBringToLife.Caption <> 'Bring To Life!' then
  begin
   StopAnimation := True;
   MustClose := True;
   Action := caNone
  end
  else
   Action := caFree;
end;

end.
